import logging
import numpy as np
import warnings

from mne.io.constants import FIFF


logger = logging.getLogger("meegsim")


def combine_stcs(stc1, stc2):
    """
    Combines the data two SourceEstimate objects. If a vertex is present in both
    stcs (e.g., as a source of 1/f noise in one and oscillation in the other),
    the corresponding signals are summed.

    Parameters
    ----------
    stc1: SourceEstimate
        First object.

    stc2: SourceEstimate
        Second object.

    Returns
    -------
    stc: SourceEstimate
        The resulting stc that contains all vertices and data from stc1 and stc2.
        If a vertex is present in both stcs, the corresponding signals are summed.
    """

    # Accumulate positions in stc1.data where time series from stc2.data
    # should be inserted
    inserters = list()

    # Keep track of the offset in stc.data while iterating over hemispheres
    offsets_old = [0]
    offsets_new = [0]

    stc = stc1.copy()
    new_data = stc2.data.copy()
    for vi, (v_old, v_new) in enumerate(zip(stc.vertices, stc2.vertices)):
        v_common, ind1, ind2 = np.intersect1d(v_old, v_new, return_indices=True)
        if v_common.size > 0:
            # Sum up signals for vertices common to stc1 and stc2
            ind1 = ind1 + offsets_old[-1]
            ind2 = ind2 + offsets_new[-1]
            stc.data[ind1] += new_data[ind2]

            # Delete the common vertices from stc2 since they do not need
            # to be processed anymore
            new_data = np.delete(new_data, ind2, axis=0)
            v_new = v_new[np.isin(v_new, v_common, invert=True)]

        # Find where to insert the remaining vertices from stc2
        inds = np.searchsorted(v_old, v_new)
        stc.vertices[vi] = np.insert(v_old, inds, v_new)
        inserters += [inds.copy()]
        offsets_old += [len(v_old)]
        offsets_new += [len(v_new)]

    inds = [ii + offset for ii, offset in zip(inserters, offsets_old[:-1])]
    inds = np.concatenate(inds)
    stc.data = np.insert(stc.data, inds, new_data, axis=0)

    return stc


def normalize_variance(data):
    """
    Divide the time series by their standard deviation to make their
    variance equal to 1.

    Parameters
    ----------
    data: array, shape (n_series, n_samples)
        Time series to be normalized.

    Returns
    -------
    data_norm: array
        Normalized time series. The variance of each row is equal to 1.
    """
    # NOTE: make a copy to keep the original waveform intact
    data_norm = data.copy()
    if data_norm.ndim == 1:
        return data_norm / np.std(data_norm)

    data_norm /= np.std(data_norm, axis=-1)[:, np.newaxis]
    return data_norm


def _extract_hemi(src):
    """
    Extract a human-readable name (lh or rh) for the provided source space
    if it is a surface one.

    Parameters
    ----------
    src: dict
        The source space to process. It should be one of the elements stored
        in the mne.SourceSpaces structure.

    Returns
    -------
    hemi: str or None
        'lh' and 'rh' are returned for left and right hemisphere, respectively.
        None is returned otherwise.
    """

    if "type" not in src or "id" not in src:
        raise ValueError(
            "The provided source space does not have the mandatory "
            "internal fields ('id' or 'type'). Please check the code "
            "that was used to generate and/or manipulate the src. "
            "It should not change or remove these fields."
        )

    if src["type"] != "surf":
        return None

    if src["id"] == FIFF.FIFFV_MNE_SURF_LEFT_HEMI:
        return "lh"

    if src["id"] == FIFF.FIFFV_MNE_SURF_RIGHT_HEMI:
        return "rh"

    raise ValueError(
        "Unexpected ID for the provided surface source space. "
        "Please check the code that was used to generate and/or "
        "manipulate the src, it should not change the 'id' field."
    )


def get_sfreq(times):
    """
    Calculate the sampling frequency of a sequence of time points.

    Parameters
    ----------
    times: ndarray
        A sequence of time points assumed to be uniformly spaced.

    Returns
    -------
    out : float
        The sampling frequency
    """

    # Check if the number of time points is less than 2
    if len(times) < 2:
        raise ValueError("The times array must contain at least two points.")

    # Calculate the differences between consecutive time points
    dt = np.diff(times)

    # Check if the mean difference is different from the first difference
    if not np.isclose(np.mean(dt), dt[0]):
        raise ValueError("Time points are not uniformly spaced.")

    return 1 / dt[0]


def unpack_vertices(vertices_lists):
    """
    Unpack a list of lists of vertices into a list of tuples.

    Parameters
    ----------
    vertices_lists : list of lists
        A list where each element is a list of vertices correspond to
        different source spaces (one or two).

    Returns
    -------
    list of tuples
        A list of tuples, where each tuple contains:
        - index: The index of the source space.
        - vertno: Vertices in corresponding source space.
    """

    if isinstance(vertices_lists, list) and not all(
        isinstance(vertices, list) for vertices in vertices_lists
    ):
        warnings.warn(
            "Input is not a list of lists. Will be assumed that there is one source space.",
            UserWarning,
        )
        vertices_lists = [vertices_lists]

    unpacked_vertices = []
    for index, vertices in enumerate(vertices_lists):
        for vertno in vertices:
            unpacked_vertices.append((index, vertno))
    return unpacked_vertices


def vertices_to_mne(vertices, src):
    """
    Convert the vertices to the MNE format (list of lists).
    """

    vertices = np.array(vertices)
    packed_vertices = [[] for _ in src]
    for src_idx in np.unique(vertices[:, 0]):
        src_vertices = vertices[vertices[:, 0] == src_idx, :]
        src_vertno = list(np.sort(src_vertices[:, 1]))
        packed_vertices[src_idx] = src_vertno

    return packed_vertices


def _hemi_to_index(hemi):
    """
    Get the index of the hemisphere (0 for lh, 1 for rh).
    """
    return ["lh", "rh"].index(hemi)


def _get_param_from_stc(stc, vertices):
    """
    Extract parameter values for specified vertices from the provided stc.

    Parameters
    ----------
    stc : mne.SourceEstimate
        The stc object that contains values for all vertices.
    vertices: list
        List of tuples (src_idx, vertno) corresponding to the vertices of interest.

    Returns
    -------
    values : array
        One value from stc for each vertex.
    """
    values = np.zeros((len(vertices),))

    # NOTE: we only support surface source estimates for now
    offsets = [0, len(stc.vertices[0])]
    for i, (src_idx, vertno) in enumerate(vertices):
        idx = offsets[src_idx] + np.searchsorted(stc.vertices[src_idx], vertno)
        values[i] = stc.data[idx]

    return values


def _get_center_of_mass(src, src_idx, vertno):
    """
    For a given set of vertices, select one that is the closest to the
    center of mass in terms of Euclidean distance.

    Parameters
    ----------
    src : SourceSpaces
        The source spaces with information about all vertices.
    src_idx : int
        The index of source space to be considered.
    vertno : list
        The list of vertex indices to be considered.

    Returns
    -------
    center_vertno : int
        The vertex from the provided list that is the closest to the center of mass.
    """
    pos = src[src_idx]["rr"][vertno, :]
    mean_pos = pos.mean(axis=0)
    dist = np.sum((pos - mean_pos) ** 2, axis=1)

    return vertno[np.argmin(dist)]



from meegsim._check import check_colors, check_scale_factors
from meegsim.sources import _get_point_sources_in_hemi, _get_patch_sources_in_hemis
from meegsim.utils import _hemi_to_index


DEFAULT_COLORS = dict(point="green", patch="Oranges", noise="black", candidate="yellow")
DEFAULT_SCALE_FACTORS = dict(point=0.75, noise=0.3, candidate=0.05)
DEFAULT_PLOT_KWARGS = dict(
    background="w",
    cortex="low_contrast",
    colorbar=False,
    clim=dict(kind="value", lims=[0, 0.5, 1]),
    transparent=True,
)


def plot_source_configuration(
    sc,
    subject,
    hemi="lh",
    colors=None,
    scale_factors=None,
    show_noise_sources=True,
    show_candidate_locations=False,
    **plot_kwargs,
):
    """
    This function can be used to plot the positions of all sources in the
    configuration. Parameters and their values are described in the docstring
    of :meth:`SourceConfiguration.plot`.
    """
    # Overwrite the default values with user input
    check_colors(colors)
    source_colors = DEFAULT_COLORS.copy()
    if colors is not None:
        source_colors.update(colors)

    check_scale_factors(scale_factors)
    source_scale_factors = DEFAULT_SCALE_FACTORS.copy()
    if scale_factors is not None:
        source_scale_factors.update(scale_factors)

    hemis = ["lh", "rh"] if hemi in ["both", "split"] else [hemi]

    # NOTE: we start with plotting all patch sources as an stc object
    # to ensure that the Brain object is initialized correctly
    patch_data_stc = _get_patch_sources_in_hemis(sc._sources.values(), sc.src, hemis)

    kwargs = DEFAULT_PLOT_KWARGS.copy()
    kwargs.update(plot_kwargs)
    brain = patch_data_stc.plot(
        subject=subject, hemi=hemi, colormap=source_colors["patch"], **kwargs
    )

    # Point/noise sources and candidate locations are added via
    # add_foci that needs to be run for each hemisphere separately
    for hemi in hemis:
        # All candidate locations (resource-heavy, disabled by default)
        if show_candidate_locations:
            src_idx = _hemi_to_index(hemi)
            candidate_locations = sc.src[src_idx]["vertno"]
            brain.add_foci(
                candidate_locations,
                coords_as_verts=True,
                hemi=hemi,
                color=source_colors["candidate"],
                scale_factor=source_scale_factors["candidate"],
            )

        # Noise sources
        if show_noise_sources:
            brain.add_foci(
                _get_point_sources_in_hemi(sc._noise_sources.values(), hemi),
                coords_as_verts=True,
                hemi=hemi,
                color=source_colors["noise"],
                scale_factor=source_scale_factors["noise"],
            )

        # Point sources (always shown)
        brain.add_foci(
            _get_point_sources_in_hemi(sc._sources.values(), hemi),
            coords_as_verts=True,
            hemi=hemi,
            color=source_colors["point"],
            scale_factor=source_scale_factors["point"],
        )

    return brain



import numpy as np

from .snr import amplitude_adjustment_factor
from .waveform import white_noise


def _prepare_sensor_noise(raw, times, random_state):
    n_chans = len(raw.ch_names)
    noise = white_noise(n_chans, times, random_state=random_state)

    # Scale the noise to equalize the mean sensor-space variance of
    # brain activity and sensor noise
    signal = raw.get_data()
    signal_var = np.trace(signal @ signal.T)
    noise_var = np.trace(noise @ noise.T)
    factor = amplitude_adjustment_factor(signal_var, noise_var, target_snr=1)

    # Dividing here since we are adjusting noise, not the signal
    noise /= factor

    return noise


def _adjustment_factors(noise_level):
    return np.sqrt(1 - noise_level), np.sqrt(noise_level)


def _adjust_sensor_noise(raw, noise, sensor_noise_level):
    """
    Mix brain activity and sensor noise together
    sensor_noise_level = power_sensor_noise / power_total
    """

    factor_signal, factor_noise = _adjustment_factors(sensor_noise_level)

    raw_mixed = raw.copy()
    raw_mixed._data = factor_signal * raw._data + factor_noise * noise

    return raw_mixed



"""
Methods for setting the coupling between two signals
"""

import numpy as np

from scipy.stats import vonmises
from scipy.signal import butter, filtfilt, hilbert

from meegsim._check import check_numeric, check_option
from meegsim.snr import get_variance, amplitude_adjustment_factor
from meegsim.utils import normalize_variance
from meegsim.waveform import narrowband_oscillation, white_noise


def _get_envelope(waveform, envelope, sfreq, fmin=None, fmax=None, random_state=None):
    check_option(
        "the amplitude envelope of the coupled waveform", envelope, ["same", "random"]
    )
    if not np.iscomplexobj(waveform):
        waveform = hilbert(waveform)

    if envelope == "same":
        return np.abs(waveform)

    if fmin is None or fmax is None:
        raise ValueError(
            "Frequency limits are required for generating the envelope of the coupled waveform"
        )
    times = np.arange(waveform.size) / sfreq
    random_waveform = narrowband_oscillation(
        1, times, fmin=fmin, fmax=fmax, random_state=random_state
    )
    random_waveform = hilbert(random_waveform)

    # TODO: here we could also mix original and random envelope with different
    # values of SNR to achieve smooth control over the resulting envelope correlation
    return np.abs(random_waveform)


def ppc_constant_phase_shift(
    waveform,
    sfreq,
    phase_lag,
    fmin=None,
    fmax=None,
    envelope="random",
    m=1,
    n=1,
    random_state=None,
):
    """
    Generate a time series that is phase coupled to the input time series with
    a constant phase lag.

    This function can be used to set up both within-frequency (1:1, default) and
    cross-frequency (n:m) coupling.

    .. note::
        This function is using Hilbert transform for manipulating the phase of
        the time series, so the result might not be meaningful if applied to
        broadband data.

    Parameters
    ----------
    waveform : array
        The input signal to be processed. It can be a real or complex time series.

    sfreq : float
        Sampling frequency of the signal, in Hz.

    phase_lag : float
        Constant phase lag to apply to the waveform in radians.

    envelope : str, {"same", "random"}
        Controls the amplitude envelope of the coupled waveform to be either randomly
        generated (default) or to be the same as the envelope of the input waveform.

    fmin : float, optional
        Lower cutoff frequency for the oscillation that gives rise to the random
        amplitude envelope (only if the ``envelope`` is set to ``"random"``).

    fmax : float, optional
        Upper cutoff frequency for the oscillation that gives rise to the random
        amplitude envelope (only if the ``envelope`` is set to ``"random"``).

    m : float, optional
        Multiplier for the base frequency of the output oscillation, default is 1.

    n : float, optional
        Multiplier for the base frequency of the input oscillation, default is 1.

    random_state : None, optional
        Random state can be fixed to provide reproducible results if the envelope
        is generated randomly. If not set, the results may differ between function calls.

    Returns
    -------
    out : array, shape (n_times,)
        The phase-coupled waveform with the same amplitude envelope as the input one.
    """
    if not np.iscomplexobj(waveform):
        waveform = hilbert(waveform)

    waveform_amp = _get_envelope(waveform, envelope, sfreq, fmin, fmax, random_state)
    waveform_angle = np.angle(waveform)
    waveform_coupled = np.real(
        waveform_amp * np.exp(1j * m / n * waveform_angle + 1j * phase_lag)
    )
    if envelope == "same":
        return normalize_variance(waveform_coupled)

    # NOTE: if the envelope was modified, we filter the result again in the target
    # frequency range to suppress possible distortions due to merging amplitude
    # envelope and phase from different time series
    b, a = butter(
        N=2, Wn=np.array([m / n * fmin, m / n * fmax]) / sfreq * 2, btype="bandpass"
    )
    waveform_coupled = filtfilt(b, a, waveform_coupled)

    return normalize_variance(waveform_coupled)


def ppc_von_mises(
    waveform,
    sfreq,
    phase_lag,
    kappa,
    fmin,
    fmax,
    envelope="random",
    m=1,
    n=1,
    random_state=None,
):
    """
    Generate a time series that is phase coupled to the input time series with
    a probabilistic phase lag based on the von Mises distribution.

    This function can be used to set up both within-frequency (1:1, default) and
    cross-frequency (n:m) coupling.

    .. note::
        This function is using Hilbert transform for manipulating the phase of
        the time series, so the result might not be meaningful if applied to
        broadband data.

    Parameters
    ----------
    waveform : array
        The input signal to be processed. It can be a real or complex time series.

    sfreq : float
        Sampling frequency (in Hz).

    phase_lag : float
        Average phase lag to apply to the waveform in radians.

    kappa : float
        Concentration parameter of the von Mises distribution. With higher kappa,
        phase shifts between input and output waveforms are more concentrated
        around the mean value provided in ``phase_lag``. With lower kappa, phase
        shifts will vary substantially for different time points.

    fmin: float
        Lower cutoff frequency of the base frequency harmonic (in Hz).

    fmax: float
        Upper cutoff frequency of the base frequency harmonic (in Hz).

    envelope : str, {"same", "random"}
        Controls the amplitude envelope of the coupled waveform to be either randomly
        generated (default) or to be the same as the envelope of the input waveform.

    m : float, optional
        Multiplier for the base frequency of the output oscillation, default is 1.

    n : float, optional
        Multiplier for the base frequency of the input oscillation, default is 1.

    random_state : None (default) or int
        Seed for the random number generator. If None (default), results will vary
        between function calls. Use a fixed value for reproducibility.

    Returns
    -------
    out : array, shape (n_times,)
        The phase-coupled waveform with the same amplitude envelope as the input one.
    """

    if not np.iscomplexobj(waveform):
        waveform = hilbert(waveform)

    waveform_amp = _get_envelope(waveform, envelope, sfreq, fmin, fmax, random_state)
    waveform_angle = np.angle(waveform)
    n_samples = waveform.size

    ph_distr = vonmises.rvs(
        kappa, loc=phase_lag, size=n_samples, random_state=random_state
    )
    waveform_coupled = np.real(
        waveform_amp * np.exp(1j * m / n * waveform_angle + 1j * ph_distr)
    )

    # NOTE: we filter the result again in the target frequency range to suppress
    # possible distortions due to separate adjustment of the phase and amplitude
    # of the coupled time series
    b, a = butter(
        N=2, Wn=np.array([m / n * fmin, m / n * fmax]) / sfreq * 2, btype="bandpass"
    )
    waveform_coupled = filtfilt(b, a, waveform_coupled)

    return normalize_variance(waveform_coupled)


def _shifted_copy_with_noise(
    waveform, sfreq, phase_lag, snr, fmin, fmax, band_limited, random_state
):
    """
    Generate a coupled time series by (1) applying a constant phase shift to the input
    waveform and (2) mixing it with noise to achieve a desired level of signal-to-noise
    ratio, which determines the resulting phase-phase and amplitude-amplitude coupling.
    """
    shifted_waveform = ppc_constant_phase_shift(
        waveform, sfreq, phase_lag, envelope="same"
    )
    signal_var = get_variance(shifted_waveform, sfreq, fmin, fmax, filter=True)

    # NOTE: to make coupling band-limited (substantial only in the band of interest),
    # we need to corrupt the rest of the coherence spectra with white noise,
    # affecting other parts of the signal apart from the frequency band of interest.
    #
    # For oscillations as our main case this is not a big deal but might be important
    # for other signals. If we filter the added noise in the frequency band of
    # interest, it leads to flat connectivity spectra but only affects target frequencies
    times = np.arange(waveform.size) / sfreq
    if band_limited:
        noise_waveform = white_noise(n_series=1, times=times, random_state=random_state)
    else:
        noise_waveform = narrowband_oscillation(
            n_series=1, times=times, fmin=fmin, fmax=fmax, random_state=random_state
        )
    noise_var = get_variance(noise_waveform, sfreq, fmin, fmax, filter=True)

    # Process the corner cases
    # SNR = inf <-> coherence = 1 -> return copy of the input
    # SNR = 0 <-> coherence = 0 -> return the generated noise
    if np.isinf(snr):
        return shifted_waveform
    if np.isclose(snr, 0):
        return noise_waveform

    factor = amplitude_adjustment_factor(signal_var, noise_var, snr)
    coupled_waveform = factor * shifted_waveform + noise_waveform

    return normalize_variance(coupled_waveform)


def _get_required_snr(coh, band_limited):
    """
    Calculate the value of SNR that is required to obtain desired coherence
    between a waveform and its copy mixed with noise.
    """
    # NOTE: prevent infinite SNR to always mix some noise in case we need to make
    # the coupling band-limited
    if band_limited and np.isclose(coh, 1, atol=1e-3):
        coh = 0.999

    return np.divide(coh**2, 1 - coh**2)


def ppc_shifted_copy_with_noise(
    waveform, sfreq, phase_lag, coh, fmin, fmax, band_limited=True, random_state=None
):
    """
    Generate a time series with desired level of coherence with the provided waveform
    in a frequency band of interest.

    The time series are generated by (1) applying a constant phase shift to the input
    waveform and (2) mixing it with a specific amount of narrowband noise. This
    function only supports within-frequency coupling.

    Parameters
    ----------
    waveform : array
        The input signal to be processed.

    sfreq : float
        Sampling frequency (in Hz).

    phase_lag : float
        Average phase lag to apply to the waveform in radians.

    coh : float
        The desired level of coherence between input and output time series.

    fmin : float
        Lower cutoff frequency of the frequency band of interest (in Hz).

    fmax : float
        Upper cutoff frequency of the frequency band of interest (in Hz).

    band_limited : bool
        Whether to limit coupling only to the frequency band of interest (True by
        default). If set to False, coupling will be the same for all frequencies,
        resulting in a flat connectivity spectra. However, the signal outside of the
        frequency band of interest will be modified negligibly.

    random_state : None (default) or int
        Seed for the random number generator. If None (default), results will vary
        between function calls. Use a fixed value for reproducibility.

    Returns
    -------
    out : array, shape (n_times,)
        The phase-coupled waveform.

    Notes
    -----
    The desired value of coherence and phase lags are obtained only on average across
    multiple simulations. For every individual output, coherence and phase lag might
    deviate from the desired values: The lower the requested coherence is, the higher
    is the variance of the output. For more information, see :doc:`this example
    </auto_examples/building_blocks/01_plot_coupling>`.
    """
    check_numeric("coherence", coh, bounds=(0, 1), allow_none=False)
    snr = _get_required_snr(coh, band_limited)
    return _shifted_copy_with_noise(
        waveform=waveform,
        sfreq=sfreq,
        phase_lag=phase_lag,
        snr=snr,
        fmin=fmin,
        fmax=fmax,
        band_limited=band_limited,
        random_state=random_state,
    )



"""
Classes that store all information about simulated sources.
Advantage of this approach over stc: in stc, we can only have one time series
per vertex, so if point sources coincide or patches overlap, we lose access
to the original time series.
"""

import numpy as np
import mne

from meegsim.utils import (
    vertices_to_mne,
    _extract_hemi,
    _get_center_of_mass,
    _get_param_from_stc,
    _hemi_to_index,
)


class _BaseSource:
    """
    An abstract class representing a source of activity.
    """

    kind = "base"

    def __init__(self, src_idx, waveform, std=1.0, name=""):
        # Current constraints:
        #  - one source corresponds to one waveform
        #  - all vertices belong to the same source space (e.g., hemisphere)
        self.src_idx = src_idx
        self.waveform = waveform
        self.std = std
        self.name = name

    @property
    def data(self):
        raise NotImplementedError(
            "The .data property should be implemented in a subclass."
        )

    @property
    def vertices(self):
        raise NotImplementedError(
            "The .vertices property should be implemented in a subclass."
        )

    def _check_compatibility(self, src):
        """
        Checks that the source is can be added to the provided src.

        Parameters
        ----------
        src: mne.SourceSpaces
            The source space where the source should be considered.

        Raises
        ------
        ValueError
            If the source does not exist in the provided src.
        """

        if self.src_idx >= len(src):
            raise ValueError(
                f"The {self.kind} source cannot be added to the provided src. "
                f"The {self.kind} source was assigned to source space {self.src_idx}, "
                f"which is not present in the provided src object."
            )

        own_vertno = [self.vertno] if self.kind == "point" else self.vertno
        missing_vertno = set(own_vertno) - set(src[self.src_idx]["vertno"])
        if missing_vertno:
            report_missing = ", ".join([str(v) for v in missing_vertno])
            raise ValueError(
                f"The {self.kind} source cannot be added to the provided src. "
                f"The source space with index {self.src_idx} does not "
                f"contain the following vertices: {report_missing}"
            )

    def to_label(self, src):
        """
        Get an mne.Label object containing all vertices belonging to the current
        source.

        Parameters
        ----------
        src : SourceSpaces
            The source space where the source should be considered.

        Returns
        -------
        label : Label
            The constructed label.
        """
        self._check_compatibility(src)

        # Make sure that we can turn the source into a label
        if src[self.src_idx]["type"] != "surf":
            raise ValueError(
                "Only sources in surface source spaces can be converted into a label"
            )

        # NOTE: we need to sort vertices before constructing the label
        vertno = np.atleast_1d(np.sort(self.vertices[:, 1]))
        return mne.Label(
            vertices=vertno,
            pos=src[self.src_idx]["rr"][vertno, :],
            # XXX: we should pass weights here once we implement patches with
            # amplitude decay
            values=np.ones_like(vertno),
            hemi="rh" if self.src_idx else "lh",
            name=self.name,
        )

    def to_stc(self, src, tstep, subject=None):
        """
        Convert the source into a SourceEstimate object in the context
        of the provided SourceSpaces.

        Parameters
        ----------
        src : SourceSpaces
            The source space where the source should be considered.
        tstep : float
            The sampling interval of the source time series (1 / sfreq).
        subject : str or None, optional
            Name of the subject that the stc corresponds to.
            If None, the subject name from the provided src is used if present.

        Returns
        -------
        stc : SourceEstimate
            SourceEstimate that corresponds to the source in the provided src.

        Raises
        ------
        ValueError
            If the source does not exist in the provided src.
        """

        self._check_compatibility(src)

        # Resolve the subject name as done in MNE
        if subject is None:
            subject = src[0].get("subject_his_id", None)

        # Convert the vertices to MNE format and construct the stc
        vertices = vertices_to_mne(self.vertices, src)
        return mne.SourceEstimate(
            data=self.data, vertices=vertices, tmin=0, tstep=tstep, subject=subject
        )


class PointSource(_BaseSource):
    """
    Point source of activity that is located in one of the vertices in
    the source space.

    Attributes
    ----------
    name : str
        The name of source.
    src_idx : int
        The index of source space that the point source belong to.
    vertno : int
        The vertex that the point source correspond to
    waveform : array
        The waveform of source activity.
    std : float, optional
        The standard deviation of the source activity (1 by default).
    hemi : str or None, optional
        Human-readable name of the hemisphere (e.g, lh or rh).
    """

    kind = "point"

    def __init__(self, name, src_idx, vertno, waveform, std=1.0, hemi=None):
        super().__init__(src_idx, waveform, std, name)

        self.vertno = vertno
        self.hemi = hemi

    def __repr__(self):
        # Use human readable names of hemispheres if possible
        src_desc = self.hemi if self.hemi else f"src[{self.src_idx}]"
        return f"<PointSource | {self.name} | {src_desc} | {self.vertno}>"

    @property
    def data(self):
        return np.atleast_2d(self.waveform)

    @property
    def vertices(self):
        return np.atleast_2d(np.array([self.src_idx, self.vertno]))

    @classmethod
    def _create(
        cls, src, times, n_sources, location, waveform, stds, names, random_state=None
    ):
        """
        This function creates point sources according to the provided input.
        """

        # Get the list of vertices (directly from the provided input or through the function)
        vertices = (
            location(src, random_state=random_state) if callable(location) else location
        )
        if len(vertices) != n_sources:
            raise ValueError("The number of sources in location does not match")

        # Get the corresponding number of time series
        data = (
            waveform(n_sources, times, random_state=random_state)
            if callable(waveform)
            else waveform
        )
        if data.shape[0] != n_sources:
            raise ValueError("The number of sources in waveform does not match")
        if data.shape[1] != len(times):
            raise ValueError("The number of samples in waveform does not match")

        # Get the std values if an stc was provided
        if isinstance(stds, mne.SourceEstimate):
            stds = _get_param_from_stc(stds, vertices)

        # Create point sources and save them as a group
        sources = []
        for (src_idx, vertno), waveform, std, name in zip(vertices, data, stds, names):
            hemi = _extract_hemi(src[src_idx])
            sources.append(
                cls(
                    name=name,
                    src_idx=src_idx,
                    vertno=vertno,
                    waveform=waveform,
                    std=std,
                    hemi=hemi,
                )
            )

        return sources


class PatchSource(_BaseSource):
    """
    Patch source of activity that is located in one of the vertices in
    the source space.

    Attributes
    ----------
    name : str
        The name of source.
    src_idx : int
        The index of source space that the patch source belong to.
    vertno : list
        The vertices that the patch sources correspond to including the central vertex.
    waveform : np.array
        The waveform of source activity.
    std : float, optional
        The standard deviation of the source activity (1 by default).
    hemi : str or None, optional
        Human-readable name of the hemisphere (e.g, lh or rh).
    """

    kind = "patch"

    def __init__(self, name, src_idx, vertno, waveform, std=1.0, hemi=None):
        super().__init__(src_idx, waveform, std, name)

        self.vertno = vertno
        self.hemi = hemi

    def __repr__(self):
        # Use human readable names of hemispheres if possible
        src_desc = self.hemi if self.hemi else f"src[{self.src_idx}]"
        n_vertno = len(self.vertno)
        vertno_desc = f"{n_vertno} vertex" if n_vertno == 1 else f"{n_vertno} vertices"
        return f"<PatchSource | {self.name} | {src_desc} | {vertno_desc} >"

    @property
    def data(self):
        # NOTE: the scaling factor is introduced to make the total variance of
        # patch activity invariant to the number of vertices in the patch
        scaling_factor = 1 / np.sqrt(len(self.vertno))
        return np.tile(self.waveform, (len(self.vertno), 1)) * scaling_factor

    @property
    def vertices(self):
        return np.array([[self.src_idx, v] for v in self.vertno])

    @classmethod
    def _create(
        cls,
        src,
        times,
        n_sources,
        location,
        waveform,
        stds,
        names,
        extents,
        subject,
        subjects_dir,
        random_state=None,
    ):
        """
        This function creates patch sources according to the provided input.
        """

        # Get the list of vertices (directly from the provided input or through the function)
        vertices = (
            location(src, random_state=random_state) if callable(location) else location
        )
        if len(vertices) != n_sources:
            raise ValueError("The number of sources in location does not match")

        # Get the corresponding number of time series
        data = (
            waveform(n_sources, times, random_state=random_state)
            if callable(waveform)
            else waveform
        )
        if data.shape[0] != n_sources:
            raise ValueError("The number of sources in waveform does not match")
        if data.shape[1] != len(times):
            raise ValueError("The number of samples in waveform does not match")

        # Pick subject name from src if not provided explicitly
        if subject is None:
            subject = src[0].get("subject_his_id", None)

        # Find patch vertices
        patch_vertices = []
        patch_stds = [] if isinstance(stds, mne.SourceEstimate) else stds
        for isource, extent in enumerate(extents):
            src_idx, vertno = vertices[isource]
            vertno = vertno if isinstance(vertno, list) else [vertno]

            # Get the std values if an stc was provided
            # NOTE: for now, we always use the value that corresponds to the
            # center of the patch but we could allow the user to select another
            # vertex
            if isinstance(stds, mne.SourceEstimate):
                center_vertno = _get_center_of_mass(src, src_idx, vertno)
                std = _get_param_from_stc(stds, [(src_idx, center_vertno)])
                patch_stds.append(std)

            # Add vertices as they are if no extent provided
            if extent is None:
                patch_vertices.append(vertno)
                continue

            # Grow the patch from center otherwise
            patch = mne.grow_labels(
                subject=subject,
                seeds=vertno,
                extents=extent,
                hemis=src_idx,
                subjects_dir=subjects_dir,
            )[0]

            # Prune vertices
            patch_vertno = [
                vert for vert in patch.vertices if vert in src[src_idx]["vertno"]
            ]
            patch_vertices.append(patch_vertno)

        # Create patch sources and save them as a group
        sources = []
        for (src_idx, _), patch_vertno, waveform, std, name in zip(
            vertices, patch_vertices, data, patch_stds, names
        ):
            hemi = _extract_hemi(src[src_idx])
            sources.append(
                cls(
                    name=name,
                    src_idx=src_idx,
                    vertno=patch_vertno,
                    waveform=waveform,
                    std=std,
                    hemi=hemi,
                )
            )

        return sources


def _combine_sources_into_stc(sources, src, tstep):
    """
    Create an stc object that contains the waveforms of all provided sources.

    Parameters
    ----------
    sources: list
        The list of point or patch sources.
    src: mne.SourceSpaces
        The source space with all candidate source locations.
    tstep: float
        The sampling interval of the source time series (1 / sfreq).

    Returns
    -------
    stc: mne.SourceEstimate
        The resulting stc object that contains all sources.
    """

    # Return immediately if no sources were provided
    if not sources:
        return None

    # Collect the data and vertices from all sources first
    data = []
    vertices = []
    for s in sources:
        s._check_compatibility(src)
        data.append(s.data)
        vertices.append(s.vertices)

    # Stack the data and vertices of all sources
    data_stacked = np.vstack(data)
    vertices_stacked = np.vstack(vertices)

    # Resolve potential repetitions: if several signals apply to the same
    # vertex, they should be summed
    unique_vertices, indices = np.unique(vertices_stacked, axis=0, return_inverse=True)
    n_unique = unique_vertices.shape[0]
    n_samples = data_stacked.shape[1]

    # Place the time courses correctly accounting for repetitions
    data = np.zeros((n_unique, n_samples))
    for idx_orig, idx_new in enumerate(indices):
        data[idx_new, :] += data_stacked[idx_orig, :]

    # Convert vertices to the MNE format
    vertices = vertices_to_mne(unique_vertices, src)

    return mne.SourceEstimate(data, vertices, tmin=0, tstep=tstep)


def _get_point_sources_in_hemi(sources, hemi):
    """
    Collect the indices of vertices (vertno) for all point sources
    belonging to the provided hemisphere.

    Parameters
    ----------
    sources: list
        A list of sources that were added to the configuration.
    hemi: str
        Hemisphere (lh or rh).

    Returns
    -------
    vertno: list
        List of indices of the corresponding vertices.
    """
    src_idx = _hemi_to_index(hemi)
    return [
        s.vertno for s in sources if isinstance(s, PointSource) and s.src_idx == src_idx
    ]


def _get_patch_sources_in_hemis(sources, src, hemis):
    """
    Collect the vertices for all patch sources belonging to the provided
    hemisphere(s).

    Parameters
    ----------
    sources: list
        A list of sources that were added to the configuration.
    src: SourceSpaces
        The source space that contains all candidate locations.
    hemis: list
        The list of hemispheres to consider.

    Returns
    -------
    stc: SourceEstimate
        An stc object that contains 1 for every vertex that is included at least
        in one of the patches and a small value near zero for all other sources.
        Non-zero value is required for the subsequent stc.plot() call.
    """
    src_indices = [_hemi_to_index(hemi) for hemi in hemis]
    n_vertno = [len(s["vertno"]) for s in src]
    # NOTE: we use a small non-zero value here to avoid problems with the
    # calculation of colorbar range in mne.viz.Brain. It is chosen to be smaller
    # than the transparency threshold (0.5) so that visually it is not noticeable
    data = [np.full((n,), 0.01) for n in n_vertno]
    for s in sources:
        if not isinstance(s, PatchSource) or s.src_idx not in src_indices:
            continue

        indices = np.searchsorted(src[s.src_idx]["vertno"], s.vertno)
        data[s.src_idx][indices] = 1
    data = np.hstack(data)

    return mne.SourceEstimate(
        data=data, vertices=[s["vertno"] for s in src], tmin=0.0, tstep=1.0
    )



import networkx as nx
import numpy as np
import warnings

from ._check import check_coupling, check_option, check_numeric_array, check_snr_params
from .configuration import SourceConfiguration
from .coupling_graph import _set_coupling
from .source_groups import PointSourceGroup, PatchSourceGroup
from .snr import _adjust_snr_local, _adjust_snr_global
from .waveform import one_over_f_noise


class SourceSimulator:
    """
    This class can be used to create configurations of sources
    of activity (e.g., narrowband oscillation or 1/f noise).

    Attributes
    ----------
    src : mne.SourceSpaces
        The source space that contains all candidate source locations.
    snr_mode : {'global', 'local'}
        The desired mode for the adjustment of the signal-to-noise ratio (SNR).

        .. note::

          * If ``'global'`` (default), the total power of **all** point/patch sources is
            adjusted relative to the total power of **all** noise sources. The target
            value of SNR should be provided in the ``snr_global`` argument of
            :meth:`~meegsim.simulate.SourceSimulator.simulate`. The spatial distribution
            of source activity can be controlled using the ``std`` argument when adding
            sources.

          * If ``'local'``, the power of **each** point/patch source is adjusted relative
            to the total power of **all** noise sources. The target value(s) of SNR
            should be provided in the ``snr`` argument when adding sources to the
            simulation via either
            :meth:`~meegsim.simulate.SourceSimulator.add_point_sources` or
            :meth:`~meegsim.simulate.SourceSimulator.add_patch_sources`
    base_std : float, optional
        The source activity is scaled by the base standard deviation before
        projecting to sensor space. By default, it is equal to :math:`10^{-9}`,
        corresponding to the dipolar moment of 1 nAm.
    """

    def __init__(self, src, snr_mode="global", base_std=1e-9):
        self.src = src
        self.base_std = base_std

        # Store groups of sources that were defined with one command
        # Store 'signal' and 'noise' separately to ease the calculation of SNR
        self._source_groups = []
        self._noise_groups = []

        # Keep track of all added sources to check name conflicts
        self._sources = []

        # Store all coupling edges in a graph
        self._coupling_graph = nx.Graph()

        # Keep track whether SNR of any source should be adjusted
        # If yes, then a forward model is required for simulation
        self.is_local_snr_adjusted = False

        # Global adjustment of SNR: combined signal vs. combined noise
        # Local adjustment of SNR: each signal separately vs. combined noise
        self.snr_mode = check_option("snr_mode", snr_mode, ["global", "local"])

    def add_point_sources(
        self,
        location,
        waveform,
        snr=None,
        std=1.0,
        location_params=dict(),
        waveform_params=dict(),
        snr_params=dict(),
        names=None,
    ):
        """
        Add point sources to the simulation.

        Parameters
        ----------
        location : list or callable
            Locations of sources can be either specified directly as a list of tuples
            (index of the src, vertno) or as a function that returns such a list.
            In the first case, source locations will be the same for every configuration.
            In the second case, configurations might differ (e.g., if the function
            returns a random location).
        waveform : array or callable
            Waveforms of source activity provided either directly in an array (fixed
            for every configuration) or as a function that generates the waveforms
            (but differ between configurations if the generation is random).
        snr : None, float, or array, optional
            SNR values for the defined sources, only used if ``snr_mode`` is set to
            ``'local'``. Can be None (no adjustment of SNR), a single value
            that is used for all sources or an array with one SNR
            value per source.
        std : float, array, or SourceEstimate, optional
            Desired standard deviation of the source activity, provided via one of
            the following options:

            - a single value that applies to all sources
            - an array with one value per source
            - a :class:`~mne.SourceEstimate` object that contains values of all
              vertices of the source space. In this case, the value will be adjusted
              for each source automatically based on its location.

            This parameter can be used in combination with the global SNR
            mode to set an arbitrary spatial distribution of source activity.
            By default, 1 is used so the variance of all sources is the same.
            If the value of ``snr`` is specified, this parameter will effectively
            be ignored.
        location_params : dict, optional
            Keyword arguments that will be passed to ``location``
            if a function is provided.
        waveform_params : dict, optional
            Keyword arguments that will be passed to ``waveform``
            if a function is provided.
        snr_params : dict, optional
            Additional parameters required for the adjustment of SNR.
            Specify ``fmin`` and ``fmax`` here to define the frequency band which
            should used for calculating the SNR.
        names : list, optional
            A list of names for each source. If not specified, the names will be
            autogenerated using the format 'auto-sgN-sM', where N is the index
            of the source group, and M is the index of the source in the group.

        Returns
        -------
        names : list
            A list of (provided or autogenerated) names for each source.
        """

        next_group_idx = len(self._source_groups)
        point_sg = PointSourceGroup.create(
            self.src,
            location,
            waveform,
            snr=snr,
            std=std,
            location_params=location_params,
            waveform_params=waveform_params,
            snr_params=snr_params,
            names=names,
            group=f"sg{next_group_idx}",
            existing=self._sources,
        )

        # Store the source group and source names
        self._source_groups.append(point_sg)
        self._sources.extend(point_sg.names)

        # Check if SNR should be adjusted
        if point_sg.snr is not None:
            if self.snr_mode == "local":
                self.is_local_snr_adjusted = True
            else:
                warnings.warn(
                    "Ignoring the provided value of local SNR since global adjustment "
                    "is enabled. To enable the local adjustment, set snr_mode to 'local' "
                    "when initializing the SourceSimulator."
                )

        # Return the names of newly added sources
        return point_sg.names

    def add_patch_sources(
        self,
        location,
        waveform,
        snr=None,
        std=1.0,
        location_params=dict(),
        waveform_params=dict(),
        snr_params=dict(),
        extents=None,
        subject=None,
        subjects_dir=None,
        names=None,
    ):
        """
        Add patch sources to the simulation.

        Parameters
        ----------
        location : list or callable
            Locations of sources can be either specified directly as a list of tuples
            (index of the src, vertno) or as a function that returns such a list.
            In the first case, source locations will be the same for every configuration.
            In the second case, configurations might differ (e.g., if the function
            returns a random location).

            Depending on the value of the ``extents`` argument, locations are
            interpreted either as set of vertices belonging to the patch (default)
            or centers of patches.
        waveform : array or callable
            Waveforms of source activity provided either directly in an array (fixed
            for every configuration) or as a function that generates the waveforms
            (might differ between configurations if the generation is random).
            For each vertex in the patch, the same waveform is currently used.
        snr : None (default), float, or array
            SNR values for the defined sources, only used if ``snr_mode`` is set to
            ``'local'``. Can be None (no adjustment of SNR, default),
            a single value that is used for all sources or an array
            with one SNR value per source.
        std : float, array, or SourceEstimate, optional
            Desired standard deviation of the **total** source activity of the
            patch (invariant to the number of vertices in the patch), provided via
            one of the following options:

            - a single value that applies to all sources
            - an array with one value per source
            - a :class:`~mne.SourceEstimate` object that contains values of all
              vertices of the source space. In this case, the value will be adjusted
              for each source automatically based on the location of its center of mass.

            This parameter can be used in combination with the global SNR
            mode to set an arbitrary spatial distribution of source activity.
            By default, 1 is used so the variance of all sources is the same.
            If the value of ``snr`` is specified, this parameter will effectively
            be ignored.
        location_params : dict, optional
            Keyword arguments that will be passed to ``location`` if a
            function is provided.
        waveform_params : dict, optional
            Keyword arguments that will be passed to ``waveform`` if a
            function is provided.
        snr_params : dict, optional
            Additional parameters required for the adjustment of SNR.
            Specify ``fmin`` and ``fmax`` here to define the frequency band which
            should used for calculating the SNR.
        extents : None (default), float, or list
            Extents (radius, in mm) of each patch. If None (default), location must
            contain all vertices belonging to the patch(es). If specified, patch are
            grown (using :func:`mne.grow_labels`) from vertices specified in
            location according to the provided values of extent. If a single number
            is provided, all patch sources have the same extent.
        subject : str, optional
            Subject name, only used when growing patch sources from the central vertex.
            If None (default), it is derived from the ``src`` object provided when
            initializing the simulator.
        subject_dir : str, optional
            Path to the directory with FreeSurfer output, only used when growing patch
            sources from the central vertex. If None (default), it is resolved automatically
            by MNE-Python. Provide the path explicitly if errors arise.
        names : list, optional
            A list of names for each source. If not specified, the names will be
            autogenerated using the format 'auto-sgN-sM', where N is the index
            of the source group, and M is the index of the source in the group.

        Returns
        -------
        names : list
            A list of (provided or autogenerated) names for each source
        """

        next_group_idx = len(self._source_groups)
        patch_sg = PatchSourceGroup.create(
            self.src,
            location,
            waveform,
            snr=snr,
            std=std,
            location_params=location_params,
            waveform_params=waveform_params,
            snr_params=snr_params,
            extents=extents,
            subject=subject,
            subjects_dir=subjects_dir,
            names=names,
            group=f"sg{next_group_idx}",
            existing=self._sources,
        )

        # Store the source group and source names
        self._source_groups.append(patch_sg)
        self._sources.extend(patch_sg.names)

        # Check if SNR should be adjusted
        if patch_sg.snr is not None:
            if self.snr_mode == "local":
                self.is_local_snr_adjusted = True
            else:
                warnings.warn(
                    "Ignoring the provided value of local SNR since global adjustment "
                    "is enabled. To enable the local adjustment, set snr_mode to 'local' "
                    "when initializing the SourceSimulator."
                )

        # Return the names of newly added sources
        return patch_sg.names

    def add_noise_sources(
        self,
        location,
        waveform=one_over_f_noise,
        std=1.0,
        location_params=dict(),
        waveform_params=dict(),
    ):
        """
        Add noise sources to the simulation. If an adjustment of SNR is needed at
        some point, these sources will be considered as noise.

        Parameters
        ----------
        location : list or callable
            Locations of sources can be either specified directly as a list of tuples
            (index of the src, vertno) or as a function that returns such a list.
            In the first case, source locations will be the same for every configuration.
            In the second case, configurations might differ (e.g., if the function
            returns a random location).
        waveform : array or callable
            Waveform provided either directly as an array or as a function.
            By default, 1/f noise with the slope of 1 is used for all noise sources.
        std : float, array, or SourceEstimate, optional
            Desired standard deviation of the source activity, provided via one of
            the following options:

            - a single value that applies to all sources
            - an array with one value per source
            - a :class:`~mne.SourceEstimate` object that contains values of all
              vertices of the source space. In this case, the value will be adjusted
              for each source automatically based on its location.

            By default, 1 is used so the variance of all noise sources is
            the same.
        location_params : dict, optional
            Keyword arguments that will be passed to ``location`` if a
            function is provided.
        waveform_params : dict, optional
            Keyword arguments that will be passed to ``waveform`` if a
            function is provided.

        Returns
        -------
        names : list
            Autogenerated names for the noise sources. The format is 'auto-ngN-sM',
            where N is the index of the noise source group, and M is the index
            of the source in the group.

        Notes
        -----
        Noise patches are currently not supported.
        """

        next_group_idx = len(self._noise_groups)
        noise_sg = PointSourceGroup.create(
            self.src,
            location,
            waveform,
            snr=None,
            std=std,
            location_params=location_params,
            waveform_params=waveform_params,
            snr_params=dict(),
            names=None,
            group=f"ng{next_group_idx}",
            existing=self._sources,
        )

        # Store the new source group and source names
        self._noise_groups.append(noise_sg)
        self._sources.extend(noise_sg.names)

        # Return the names of newly added sources
        return noise_sg.names

    def set_coupling(self, coupling, **common_params):
        """
        Set coupling between sources that were added to the simulator.

        Parameters
        ----------
        coupling : tuple or dict
            Provide a tuple ``(u, v)`` to define one pair of coupled sources
            or a dictionary to define multiple coupling edges at once. ``u`` and
            ``v`` are the names of sources that should be coupled. Both
            sources should be added to the simulation prior to setting the coupling.

            If used, the dictionary should contain tuples ``(u, v)`` as keys,
            while the values should be dictionaries with keyword arguments
            of the coupling method. Use this dictionary to define coupling
            parameters that are specific for a given edge. Such definitions will
            also override the common parameters (described below).
        **common_params : dict, optional
            Additional coupling parameters that apply to each edge defined in the
            coupling dictionary or the single edge if a tuple was provided.

        Notes
        -----
        For the information on required coupling parameters, please refer to the
        :doc:`documentation </api/coupling>` of the corresponding coupling method(s).

        Examples
        --------
        Adding a single connectivity edge:

        >>> from meegsim.coupling import ppc_von_mises
        ...
        ... sim.set_coupling(('s1', 's2'), method=ppc_von_mises,
        ...                  kappa=1, phase_lag=0, fmin=8, fmax=12)

        Adding multiple connectivity edges at once:

        >>> from meegsim.coupling import ppc_von_mises
        ...
        ... sim.set_coupling(coupling={
        ...     ('s1', 's2'): dict(kappa=1, phase_lag=np.pi/3, fmin=10),
        ...     ('s2', 's3'): dict(kappa=0.5, phase_lag=-np.pi/6)
        ... }, method=ppc_von_mises, fmin=8, fmax=12)

        In the example above, ``method`` and ``fmax`` values apply to both
        coupling edges, while ``kappa`` and ``phase_lag`` are edge-specific.
        ``fmin`` is defined as a common parameter but also has a different
        value for the edge ``('s1', 's2')``. Therefore, it will be set to `6`
        for the edge ``('s1', 's2')`` and to `8` for the edge ``('s2', 's3')``.
        """

        # Convert tuple to a dictionary with empty coupling params
        if isinstance(coupling, tuple):
            coupling = {coupling: dict()}

        for coupling_edge, coupling_params in coupling.items():
            params = check_coupling(
                coupling_edge,
                coupling_params,
                common_params,
                self._sources,
                self._coupling_graph,
            )

            # Add the coupling edge
            source, target = coupling_edge
            self._coupling_graph.add_edge(source, target, **params)

    def simulate(
        self,
        sfreq,
        duration,
        fwd=None,
        snr_global=None,
        snr_params=dict(),
        random_state=None,
    ):
        """
        Simulate a configuration of defined sources.

        Parameters
        ----------
        sfreq : float
            The sampling frequency of the simulated data, in Hz.
        duration : float
            Duration of the simulated data, in seconds.
        fwd : None or Forward, optional
            The forward model, only to be used for the adjustment of SNR.
            If no adjustment is performed, the forward model is not required.
        snr_global : float or None, optional
            The value of global SNR, only used if the ``snr_mode`` is set to
            ``'global'``. If None (default), no adjustment of global SNR is performed.
        snr_params : dict, optional
            Additional parameters required for the adjustment of global SNR.
            Specify ``fmin`` and ``fmax`` here to define the frequency band which
            should used for calculating the SNR.
        random_state : int or None, optional
            The random state can be provided to obtain reproducible configurations.
            If None (default), the simulated data will differ between function calls.

        Returns
        -------
        sc : SourceConfiguration
            The source configuration, which contains the defined sources and
            their corresponding waveforms.
        """

        if not (self._source_groups or self._noise_groups):
            raise ValueError("No sources were added to the configuration.")

        # We expect None or one value that applies to all sources
        snr_global = check_numeric_array(
            "global SNR", snr_global, n_sources=1, bounds=(0, None), allow_none=True
        )
        snr_params = check_snr_params(snr_params, snr_global)

        # Check the forward model and auto-fill info if needed
        is_global_snr_adjusted = self.snr_mode == "global" and snr_global is not None
        is_local_snr_adjusted = self.snr_mode == "local" and self.is_local_snr_adjusted
        if snr_global is not None and self.snr_mode == "local":
            warnings.warn(
                "Ignoring the provided value of global SNR since local adjustment "
                "is enabled. To enable the global adjustment, set snr_mode to 'global' "
                "when initializing the SourceSimulator."
            )
        if (is_global_snr_adjusted or is_local_snr_adjusted) and fwd is None:
            raise ValueError("A forward model is required for the adjustment of SNR.")

        # Initialize the SourceConfiguration
        sc = SourceConfiguration(self.src, sfreq, duration, random_state=random_state)

        # Simulate signal and noise
        sources, noise_sources = _simulate(
            source_groups=self._source_groups,
            noise_groups=self._noise_groups,
            coupling_graph=self._coupling_graph,
            snr_mode=self.snr_mode,
            snr_global=snr_global,
            snr_params=snr_params,
            is_local_snr_adjusted=self.is_local_snr_adjusted,
            src=self.src,
            times=sc.times,
            fwd=fwd,
            base_std=self.base_std,
            random_state=random_state,
        )

        # Add the sources to the simulated configuration
        sc._sources = sources
        sc._noise_sources = noise_sources

        return sc


def _simulate(
    source_groups,
    noise_groups,
    coupling_graph,
    snr_mode,
    snr_global,
    snr_params,
    is_local_snr_adjusted,
    src,
    times,
    fwd,
    base_std,
    random_state=None,
):
    """
    This function describes the simulation workflow.
    """

    # Generate unique random states for each simulation step:
    #  - generation of noise source groups
    #  - generation of point/patch source groups
    #  - generation of coupling
    # NOTE: if we don't perform this, then callable-based sources defined in
    # different calls will have identical locations and waveforms
    is_coupling_required = coupling_graph.number_of_edges() > 0
    n_seeds = len(noise_groups) + len(source_groups) + is_coupling_required
    seeds = list(np.random.SeedSequence(random_state).generate_state(n_seeds))

    # Simulate all sources independently first (no coupling yet)
    noise_sources = []
    for ng in noise_groups:
        noise_sources.extend(ng.simulate(src, times, random_state=seeds.pop(0)))
    noise_sources = {s.name: s for s in noise_sources}

    sources = []
    for sg in source_groups:
        sources.extend(sg.simulate(src, times, random_state=seeds.pop(0)))
    sources = {s.name: s for s in sources}

    # Setup the desired coupling patterns
    # The time courses are changed for some of the sources in the process
    if is_coupling_required:
        _set_coupling(sources, coupling_graph, times, random_state=seeds.pop(0))

    # Set the standard deviation of all sources w.r.t. base std
    # NOTE: this should also be helpful to get less warnings about unreasonably
    # high values of source activity from apply_forward_raw
    for s in sources.values():
        s.waveform *= base_std * s.std
    for s in noise_sources.values():
        s.waveform *= base_std * s.std

    # Adjust the SNR if needed
    if snr_mode == "global" and snr_global is not None:
        tstep = times[1] - times[0]
        _adjust_snr_global(
            src, fwd, snr_global, snr_params, tstep, sources, noise_sources
        )
    elif is_local_snr_adjusted:
        tstep = times[1] - times[0]
        _adjust_snr_local(src, fwd, tstep, sources, source_groups, noise_sources)

    return sources, noise_sources



import networkx as nx
import numpy as np

from .utils import get_sfreq


def traverse_tree(tree, start_node=None, random_state=None):
    """
    Generate a list of walkaround paths in a tree starting from start_node.

    Walkaround paths are pairs of nodes where each pair represents an edge
    in the tree, starting from the specified start_node.

    Parameters:
    ----------
    tree : networkx.Graph
        The tree in which to generate walkaround paths.
    start_node : int
        The node from which to start generating paths.
        If start_node is None (default), the start node will be drawn randomly.
    random_state : int or None, optional
        Seed for the random number generator. If start_node is None (default), the
        start node will be drawn randomly, and results will vary between function calls.

    Returns:
    -------
    out : list of tuples
        A list of pairs of nodes representing walkaround paths.
    """

    if start_node is None:
        # take random
        rng = np.random.default_rng(random_state)
        start_node = rng.choice(list(tree.nodes))

    return list(nx.dfs_edges(tree, source=start_node))


def generate_walkaround(coupling_graph, random_state=None):
    """
    Constructs a graph from the provided edge list and attributes, and identifies walkaround paths in tree topologies.

    Parameters
    ----------
    coupling_graph : nx.Graph
        The coupling graph that describes the desired connectivity patterns.
        All edges should have the coupling parameters as attributes.
    random_state : int or None, optional
        Seed for the random number generator. If start_node is None, the start node will be drawn
        randomly, and results will vary between function calls. default = None.

    Returns
    -------
    walkaround : list of tuples
        A list of coupling edges (source, target) ordered in a way that guarantees the
        desired coupling for all the edges.
    """

    if not nx.is_forest(coupling_graph):
        raise ValueError("The graph contains cycles. Cycles are not supported.")

    # iterate over connected components
    walkaround = []
    for component in nx.connected_components(coupling_graph):
        subgraph = coupling_graph.subgraph(component)

        # build the path starting from random node
        walkaround_paths = traverse_tree(
            subgraph, start_node=None, random_state=random_state
        )
        walkaround.extend(walkaround_paths)

    return walkaround


def _set_coupling(sources, coupling_graph, times, random_state):
    """
    This function traverses the coupling graph and executes the simulation
    of coupling for each edge in the graph.

    Parameters
    ----------
    sources : dict
        Simulated sources. Their waveforms are modified in-place by the
        coupling function(s).
    coupling_graph : nx.Graph
        The coupling graph that describes the desired connectivity pattern.
    times : array-like
        The time points for all samples in the waveform.
    random_state : int or None
        The random state that could be fixed to ensure reproducibility.
    """
    # Traverse the graph to ensure that coupling is set correctly
    walkaround = generate_walkaround(coupling_graph, random_state=random_state)

    # Generate random states for each edge
    # NOTE: if we don't perform this, same noise or phase lag distribution might be
    # used for different coupling edges, which is most likely not desirable
    n_edges = len(walkaround)
    seeds = list(np.random.SeedSequence(random_state).generate_state(n_edges))

    for name1, name2 in walkaround:
        # Get the sources by their names
        s1, s2 = sources[name1], sources[name2]

        # Get the corresponding coupling parameters
        coupling_params = coupling_graph.get_edge_data(name1, name2)

        # Extract the coupling method temporarily
        tmp_coupling_params = coupling_params.copy()
        coupling_fn = tmp_coupling_params.pop("method")

        # Adjust the waveform of s2 to be coupled with s1
        s2.waveform = coupling_fn(
            s1.waveform,
            get_sfreq(times),
            **tmp_coupling_params,
            random_state=seeds.pop(0),
        )



"""
This classes store the information about source groups that were
defined by the user until we actually start simulating the data.
"""

import mne

from meegsim._check import (
    check_location,
    check_waveform,
    check_numeric_array,
    check_snr_params,
    check_stc_as_param,
    check_names,
    check_extents,
)
from meegsim.sources import PointSource, PatchSource


def generate_names(group, n_sources):
    """
    Automatically generate names for sources belonging to the group.

    Parameters
    ----------
    group: str
        The name of the source group.
    n_sources: int
        The number of sources in the group.

    Returns
    -------
    names: list
        The auto-generated names in the format 'auto-G-sX', where G is the group name,
        and X is the index of the source in the group.
    """

    return [f"auto-{group}-s{idx}" for idx in range(n_sources)]


class _BaseSourceGroup:
    def simulate(self):
        raise NotImplementedError(
            "The simulate() method should be implemented in a subclass."
        )


class PointSourceGroup(_BaseSourceGroup):
    def __init__(self, n_sources, location, waveform, snr, snr_params, std, names):
        super().__init__()

        # Store the defined number of vertices to raise an error
        # if the output of location function has a different size
        self.n_sources = n_sources

        # Store the provided information
        self.location = location
        self.waveform = waveform
        self.snr = snr
        self.snr_params = snr_params
        self.std = std
        self.names = names

    def __repr__(self):
        location_desc = "list"
        if callable(self.location):
            # extract the name of the function from the partial object if possible
            location_desc = getattr(self.location.func, "__name__", "callable")
        location_desc = f"location={location_desc}"

        waveform_desc = "array"
        if callable(self.waveform):
            # extract the name of the function from the partial object if possible
            waveform_desc = getattr(self.waveform.func, "__name__", "callable")
        waveform_desc = f"waveform={waveform_desc}"

        return f"<PointSourceGroup | {self.n_sources} sources | {location_desc} | {waveform_desc}>"

    def simulate(self, src, times, random_state=None):
        return PointSource._create(
            src,
            times,
            self.n_sources,
            self.location,
            self.waveform,
            self.std,
            self.names,
            random_state=random_state,
        )

    @classmethod
    def create(
        cls,
        src,
        location,
        waveform,
        snr,
        std,
        location_params,
        waveform_params,
        snr_params,
        names,
        group,
        existing,
    ):
        """
        Check the provided input for all fields and create a source group that
        would store this information.

        Parameters
        ----------
        src: mne.SourceSpaces
            The source space that contains all candidate source locations.
        location: list or callable
            The location provided by the user.
        waveform: list of callable
            The waveform provided by the user.
        snr: None, float, or array
            The SNR values provided by the user.
        std: float, array, or mne.SourceEstimate
            The values of standard deviation provided by the user.
        location_params: dict, optional
            Additional keyword arguments for the location function.
        waveform_params: dict, optional
            Additional keyword arguments for the waveform function.
        snr_params: dict, optional
            Additional parameters for the adjustment of SNR.
        names:
            The names of sources provided by the user.
        group:
            The name of the source group. Only used for generating the names
            of sources automatically.
        existing:
            The names of sources that were already taken.

        Returns
        -------
        group: PointSourceGroup
            A source group object with checked and preprocessed user input.
        """

        # Check the user input
        location, n_sources = check_location(location, location_params, src)
        waveform = check_waveform(waveform, waveform_params, n_sources)
        snr = check_numeric_array(
            "SNR", snr, n_sources, bounds=(0, None), allow_none=True
        )
        snr_params = check_snr_params(snr_params, snr)
        if isinstance(std, mne.SourceEstimate):
            check_stc_as_param(std, src)
        else:
            std = check_numeric_array("std", std, n_sources, bounds=(0, None))

        # Auto-generate or check the provided source names
        if not names:
            names = generate_names(group, n_sources)
        else:
            check_names(names, n_sources, existing)

        return cls(n_sources, location, waveform, snr, snr_params, std, names)


class PatchSourceGroup(_BaseSourceGroup):
    def __init__(
        self,
        n_sources,
        location,
        waveform,
        snr,
        snr_params,
        std,
        extents,
        subject,
        subjects_dir,
        names,
    ):
        super().__init__()

        # Store the defined number of vertices to raise an error
        # if the output of location function has a different size
        self.n_sources = n_sources

        # Store the provided information
        self.location = location
        self.waveform = waveform
        self.snr = snr
        self.snr_params = snr_params
        self.std = std
        self.names = names
        self.extents = extents
        self.subject = subject
        self.subjects_dir = subjects_dir

    def __repr__(self):
        location_desc = "list"
        if callable(self.location):
            # extract the name of the function from the partial object if possible
            location_desc = getattr(self.location.func, "__name__", "callable")
        location_desc = f"location={location_desc}"

        waveform_desc = "array"
        if callable(self.waveform):
            # extract the name of the function from the partial object if possible
            waveform_desc = getattr(self.waveform.func, "__name__", "callable")
        waveform_desc = f"waveform={waveform_desc}"

        return f"<PatchSourceGroup | {self.n_sources} sources | {location_desc} | {waveform_desc}>"

    def simulate(self, src, times, random_state=None):
        return PatchSource._create(
            src,
            times,
            self.n_sources,
            self.location,
            self.waveform,
            self.std,
            self.names,
            self.extents,
            self.subject,
            self.subjects_dir,
            random_state=random_state,
        )

    @classmethod
    def create(
        cls,
        src,
        location,
        waveform,
        snr,
        std,
        location_params,
        waveform_params,
        snr_params,
        extents,
        subject,
        subjects_dir,
        names,
        group,
        existing,
    ):
        """
        Check the provided input for all fields and create a source group that
        would store this information.

        Parameters
        ----------
        src: mne.SourceSpaces
            The source space that contains all candidate source locations.
        location: list or callable
            The location provided by the user.
        waveform: list of callable
            The waveform provided by the user.
        snr:
            The SNR values provided by the user.
        std: float, array, or mne.SourceEstimate
            The values of standard deviation provided by the user.
        location_params: dict, optional
            Additional keyword arguments for the location function.
        waveform_params: dict, optional
            Additional keyword arguments for the waveform function.
        snr_params:
            Additional parameters for the adjustment of SNR.
        extents: list
            Extents (radius in mm) of each patch provided by the user.
        subject: str, optional
            Subject name.
        subject_dir: str, optional
            Path to the directory with FreeSurfer output.
        names:
            The names of sources provided by the user.
        group:
            The name of the source group. Only used for generating the names
            of sources automatically.
        existing:
            The names of sources that were already taken.

        Returns
        -------
        group: PatchSourceGroup
            A source group object with checked and preprocessed user input.
        """

        # Check the user input
        location, n_sources = check_location(location, location_params, src)
        waveform = check_waveform(waveform, waveform_params, n_sources)
        snr = check_numeric_array(
            "SNR", snr, n_sources, bounds=(0, None), allow_none=True
        )
        snr_params = check_snr_params(snr_params, snr)
        if isinstance(std, mne.SourceEstimate):
            check_stc_as_param(std, src)
        else:
            std = check_numeric_array("std", std, n_sources, bounds=(0, None))
        extents = check_extents(extents, n_sources)

        # Auto-generate or check the provided source names
        if not names:
            names = generate_names(group, n_sources)
        else:
            check_names(names, n_sources, existing)

        return cls(
            n_sources,
            location,
            waveform,
            snr,
            snr_params,
            std,
            extents,
            subject,
            subjects_dir,
            names,
        )



import numpy as np
import mne
import warnings

from scipy.signal import butter, filtfilt

from meegsim.sources import _combine_sources_into_stc


def get_variance(waveform, sfreq, fmin=None, fmax=None, filter=False):
    """
    Estimate the variance of time series, optionally in a frequency band of interest.

    Parameters
    ----------
    waveform : array, shape (n_times,)
        The input waveform.
    sfreq : float
        The sampling frequency, in Hz.
    fmin : float or None, optional
        Lower frequency cutoff for the frequency band of interest, in Hz.
    fmax : float or None, optional
        Upper frequency cutoff for the frequency band of interest, in Hz.
    filter : bool
        Whether filtering in a frequency band should be applied.

    Returns
    -------
    var : float
        Variance of the input waveform.
    """
    if filter:
        b, a = butter(2, np.array([fmin, fmax]) / sfreq * 2, btype="bandpass")
        waveform = filtfilt(b, a, waveform)
    return np.var(waveform)


def get_sensor_space_variance(stc, fwd, fmin=None, fmax=None, filter=False):
    """
    Estimate the sensor space variance of the provided stc

    Parameters
    ----------
    stc: mne.SourceEstimate
        Source estimate containing signal or noise (vertices x times).

    fwd: mne.Forward
        Forward model.

    fmin: float, optional
        Lower cutoff frequency (in Hz). default = None.

    fmax: float, optional
        Upper cutoff frequency (in Hz). default = None.

    filter: bool, optional
        Indicate if filtering in the band of oscillations is required. default = False.

    Returns
    -------
    stc_var: float
        Sensor space variance in the frequency band, if specified.
    """

    try:
        raw = mne.apply_forward_raw(
            fwd, stc, fwd["info"], on_missing="raise", verbose=False
        )
    except ValueError:
        raise ValueError(
            "The provided forward model does not contain some of the "
            "simulated sources, so the SNR cannot be adjusted."
        )

    raw_data = raw.get_data()
    if filter:
        if fmin is None or fmax is None:
            raise ValueError(
                "Frequency band limits are required for the adjustment of SNR."
            )

        b, a = butter(2, np.array([fmin, fmax]) / stc.sfreq * 2, btype="bandpass")
        raw_data = filtfilt(b, a, raw_data, axis=1)

    n_samples = raw_data.shape[1]
    sensor_cov = (raw_data @ raw_data.T) / (n_samples - 1)
    sensor_var = np.mean(np.diag(sensor_cov))

    return sensor_var


def amplitude_adjustment_factor(signal_var, noise_var, target_snr):
    """
    Derive the adjustment factor for signal amplitude that allows obtaining the target SNR

    Parameters
    ----------
    signal_var: float
        Variance of the simulated signal with respect to leadfield. Can be obtained with
        a function snr.get_sensor_space_variance.

    noise_var: float
        Variance of the simulated noise with respect to leadfield. Can be obtained with
        a function snr.get_sensor_space_variance.

    target_snr: float
        Value of a desired SNR for the signal.

    Returns
    -------
    factor: float
        The original signal should be multiplied by this value to obtain the desired SNR.
    """

    snr_current = np.divide(signal_var, noise_var)

    if np.isinf(snr_current):
        raise ValueError(
            "The noise variance appears to be zero, so the initial SNR "
            "cannot be calculated. Please check the created noise."
        )

    factor = np.sqrt(target_snr / snr_current)

    if np.isinf(factor):
        raise ValueError(
            "The signal variance and thus the initial SNR appear to be "
            "zero, so SNR cannot be adjusted. Please check the created "
            "signals."
        )

    return factor


def _adjust_snr_local(src, fwd, tstep, sources, source_groups, noise_sources):
    """
    Perform the adjustment of local SNR: the power of each source is adjusted
    relative to the power of all noise sources combined. The waveforms are
    adjusted in-place.
    """
    # Get the stc and leadfield of all noise sources
    if not noise_sources:
        raise ValueError(
            "No noise sources were added to the simulation, so the SNR "
            "cannot be adjusted."
        )
    stc_noise = _combine_sources_into_stc(noise_sources.values(), src, tstep)

    # Adjust the SNR of sources in each source group
    for sg in source_groups:
        if sg.snr is None:
            continue

        # Estimate the noise variance in the specified frequency band
        fmin, fmax = sg.snr_params["fmin"], sg.snr_params["fmax"]
        noise_var = get_sensor_space_variance(
            stc_noise, fwd, fmin=fmin, fmax=fmax, filter=True
        )

        # Adjust the amplitude of each source in the group to match the target SNR
        for name, target_snr in zip(sg.names, sg.snr):
            s = sources[name]

            # NOTE: taking a safer approach for now and filtering
            # even if the signal is already a narrowband oscillation
            signal_var = get_sensor_space_variance(
                s.to_stc(src, tstep), fwd, fmin=fmin, fmax=fmax, filter=True
            )

            # NOTE: patch sources might require more complex calculations
            # if the within-patch correlation is not equal to 1
            factor = amplitude_adjustment_factor(signal_var, noise_var, target_snr)
            s.waveform *= factor


def _adjust_snr_global(src, fwd, snr_global, snr_params, tstep, sources, noise_sources):
    """
    Perform the adjustment of global SNR: the power of all sources combined
    is adjusted relative to the power of all noise sources combined. The
    waveforms are adjusted in-place.
    """
    # Combine signal/noise sources
    if not sources:
        warnings.warn(
            "No point/patch sources were added to the simulation, "
            "skipping the requested adjustment of global SNR."
        )
        return

    stc_signal = _combine_sources_into_stc(sources.values(), src, tstep)

    if not noise_sources:
        raise ValueError(
            "No noise sources were added to the simulation, so the global SNR "
            "cannot be adjusted."
        )
    stc_noise = _combine_sources_into_stc(noise_sources.values(), src, tstep)

    # Get sensor-space variance of signal and noise
    fmin, fmax = snr_params["fmin"], snr_params["fmax"]
    noise_var = get_sensor_space_variance(
        stc_noise, fwd, fmin=fmin, fmax=fmax, filter=True
    )
    signal_var = get_sensor_space_variance(
        stc_signal, fwd, fmin=fmin, fmax=fmax, filter=True
    )

    # Adjust the amplitudes of all sources by the same (!) factor
    factor = amplitude_adjustment_factor(signal_var, noise_var, snr_global)
    for s in sources.values():
        s.waveform *= factor



"""
Template waveforms: narrowband oscillation, white and 1/f noise
"""

import colorednoise as cn
import numpy as np
import warnings

from scipy.signal import butter, filtfilt

from .utils import normalize_variance, get_sfreq


def narrowband_oscillation(
    n_series, times, *, fmin=None, fmax=None, order=2, random_state=None
):
    """
    Generate oscillatory-like time series by filtering white noise
    in a requested frequency band.

    Parameters
    ----------
    n_series : int
        The number of time series to generate.

    times : array
        Array of time points (each one represents time in seconds).

    fmin : None (default) or float
        Lower cutoff frequency (in Hz). If None (default), 8 Hz are used as the
        cutoff.

    fmax : None (default) or float
        Upper cutoff frequency (in Hz). If None (default), 12 Hz are used as the
        cutoff.

    order : int, optional
        The order of the filter. By default, the order is equal to 2.

    random_state : None (default) or int
        Seed for the random number generator. If None (default), results will vary
        between function calls. Use a fixed value for reproducibility.

    Returns
    -------
    out : array, shape (n_series, n_times)
        Generated filtered white noise.
    """

    if fmin is None:
        warnings.warn("fmin was None. Setting fmin to 8 Hz", UserWarning)
        fmin = 8.0
    if fmax is None:
        warnings.warn("fmax was None. Setting fmax to 12 Hz", UserWarning)
        fmax = 12.0

    if fmin >= fmax:
        raise ValueError("fmin must be smaller than fmax.")
    if fmin <= 0 or fmax <= 0:
        raise ValueError("filter frequencies must be greater than 0")

    if not isinstance(order, int) or order <= 0:
        raise ValueError("order must be a positive integer.")

    fs = get_sfreq(times)
    rng = np.random.default_rng(seed=random_state)
    data = rng.standard_normal(size=(n_series, times.size))
    b, a = butter(N=order, Wn=np.array([fmin, fmax]) / fs * 2, btype="bandpass")
    data = filtfilt(b, a, data, axis=1)
    return normalize_variance(data)


def one_over_f_noise(n_series, times, *, slope=1, random_state=None):
    """
    Generate time series of 1/f noise with desired slope.

    Parameters
    ----------
    n_series : int
        The number of time series to generate.

    times : array
        Array of time points (each one represents time in seconds). Only the size
        of the array is used for noise generation.

    slope : float, optional
        Exponent of the power-law spectrum. By default, it is equal to 1.

    random_state : None (default) or int
        Seed for the random number generator. If None (default), results will vary
        between function calls. Use a fixed value for reproducibility.

    Returns
    -------
    out : array, shape (n_series, n_times)
        Generated 1/f noise.
    """

    data = cn.powerlaw_psd_gaussian(
        slope, size=(n_series, times.size), random_state=random_state
    )
    return normalize_variance(data)


def white_noise(n_series, times, *, random_state=None):
    """
    Generate time series of white noise (e.g., to use for modeling
    measurement noise in sensor space data)

    Parameters
    ----------
    n_series : int
        Number of time series to generate.

    times : array
        Array of time points (each one represents time in seconds). Only the size
        of the array is used for noise generation.

    random_state : None (default) or int
        Seed for the random number generator. If None (default), results will vary
        between function calls. Use a fixed value for reproducibility.

    Returns
    -------
    out : array, shape (n_series, n_times)
        Generated white noise.
    """

    rng = np.random.default_rng(seed=random_state)
    data = rng.standard_normal(size=(n_series, times.size))
    return normalize_variance(data)



"""
Methods for selecting locations of the sources that accept the following arguments:
    * it would be easier to have all methods accept src as the first argument by default
    * ideally, random_state (to allow reproducibility, still need to test how it would work)

Many options are already covered by mne.simulation.select_source_in_label so we can reuse the functionality under the hood.
"""

import numpy as np

from meegsim.utils import unpack_vertices


def select_random(src, *, n=1, vertices=None, sort_output=False, random_state=None):
    """
    Randomly selects a specified number of vertices from a given source space.

    Parameters
    ----------
    src : SourceSpaces
        An instance of source spaces to select the vertices from.

    n : int, optional
        The number of random vertices to select. By default, one vertex is selected.

    vertices : None (default) or list
        A subset of vertices to choose from.
        If None (default), the function uses all vertices from the provided ``src``.

    sort_output : bool, optional
        Indicates if sorting is needed for the output. By default, the output is
        not sorted.

    random_state : None (default) or int
        Seed for the random number generator. If None (default), results will vary
        between function calls. Use a fixed value for reproducibility.

    Returns
    -------
    selected : list
        A list of tuples (index of the source space, vertno of the selected vertex).
    """
    rng = np.random.default_rng(seed=random_state)

    if len(src) not in [1, 2]:
        raise ValueError(
            "Src must contain either one (volume) or two (surface) source spaces."
        )

    src_unpacked = unpack_vertices([list(s["vertno"]) for s in src])

    vertices = unpack_vertices(vertices) if vertices else src_unpacked
    vertices_not_in_src = set(vertices) - set(src_unpacked)
    if vertices_not_in_src:
        raise ValueError("Some vertices are not contained in the src.")

    if n > len(vertices):
        raise ValueError("Number of vertices to select exceeds available vertices.")

    selected_vertno = rng.choice(vertices, size=n, replace=False)
    selected_vertno = list(map(tuple, selected_vertno))
    if sort_output:
        selected_vertno = sorted(selected_vertno)

    return selected_vertno



"""
This module contains all functions that check the provided input
for SourceSimulator:

 - location and waveform functions
 - SNR values
 - parameters for adjusting SNR
 - coupling parameters
 - source names
"""

import numpy as np
import matplotlib.pyplot as plt
import warnings

from functools import partial
from matplotlib.colors import is_color_like

from meegsim.utils import logger


VIZ_SOURCE_TYPES = ["point", "patch", "noise", "candidate"]


def check_numeric(context, value, bounds=(None, None), allow_none=True):
    if allow_none and value is None:
        return value

    try:
        value = float(value)
    except (TypeError, ValueError):  # None leads to TypeError
        raise ValueError(f"Expected {context} to be a float number, got {type(value)}")

    lo, hi = bounds
    if lo is not None and value < lo:
        raise ValueError(f"Expected {context} to be {lo} or greater")
    if hi is not None and value > hi:
        raise ValueError(f"Expected {context} to be {hi} or lower")

    return value


def check_option(context, value, allowed_values, allow_none=False):
    """
    Check that the provided value is None (if allowed) or belongs to the list
    of the allowed values.

    Parameters
    ----------
    context : str
        Context information for the error message.
    value
        The value to be checked.
    allowed_values : list
        The list of values that are allowed.
    allow_none : bool
        If True, None is also allowed.

    Returns
    -------
    value
        The provided value

    Raises
    ------
    ValueError
        If the provided value is not allowed.
    """
    if allow_none and value is None:
        return value

    if value not in allowed_values:
        raise ValueError(f"The value {value} is not allowed for {context}")

    return value


def check_callable(context, fun, *args, **kwargs):
    """
    Check whether the provided function can be run successfully.
    The function is always run with random_state set to 0 for consistency.

    Parameters
    ----------
    name: str
        The name of the function to describe the context in the error message.
    fun: callable
        The function to be checked.
    *args: tuple
        Positional arguments that need to be passed to the function.
    **kwargs: dict, optional
        Keyword arguments the need to be passed to the function.

    Returns
    -------
    result
        The result of the function if the call was successful.

    Raises
    ------
    Exception
        Any exception that occurs during the function call.
    """

    try:
        return fun(*args, **kwargs, random_state=0)
    except:
        logger.error(
            f"An error occurred when trying to call the "
            f"provided function for: {context}"
        )
        raise


def check_vertices_list_of_tuples(vertices):
    """
    Check that the provided vertices are a list of tuples corresponding
    to our format:

    [(src_idx, vertno), (src_idx, vertno), ...]

    Parameters
    ----------
    vertices
        The input to be checked.

    Raises
    ------
    ValueError
        If any violation of the format is detected.
    """

    if not isinstance(vertices, (list, tuple)):
        raise ValueError(
            f"Expected vertices to be a list or a tuple, got {type(vertices)}"
        )

    for i, el in enumerate(vertices):
        if not isinstance(el, (list, tuple)):
            raise ValueError(
                f"Expected each element of the vertices list to "
                f"be a list or a tuple, does not hold for "
                f"element {el}"
            )

        if len(el) != 2:
            raise ValueError(
                f"Expected each element of the vertices list to "
                f"contain 2 values, does not hold for element {el}"
            )


def check_vertices_in_src(vertices, src):
    """
    Checks that all vertices are present in the provided src.

    Parameters
    ----------
    vertices: list
        The vertices to be checked.
    src: mne.SourceSpaces
        The source space which should contain all vertices.

    Raises
    ------
    ValueError
        In case any vertex is not present in the provided src.
    """
    for v in vertices:
        src_idx, vertno = v
        if src_idx >= len(src):
            raise ValueError(
                f"Vertex {v} belongs to the source space {src_idx}, "
                f"which is not present in the provided src"
            )

        vertno = [vertno] if not isinstance(vertno, list) else vertno
        missing_vertno = set(vertno) - set(src[src_idx]["vertno"])
        if missing_vertno:
            report_missing = ", ".join([str(v) for v in missing_vertno])
            vertex_desc = "Vertex" if len(missing_vertno) == 1 else "Vertices"
            verb = "is" if len(missing_vertno) == 1 else "are"
            raise ValueError(
                f"{vertex_desc} {report_missing} {verb} not present in the provided "
                f"src[{src_idx}]"
            )


def check_location(location, location_params, src):
    """
    Check the user input for the location of sources.
    If location is a callable, it should not lead to an error.
    The result of the location function (if callable) or the location itself
    (if a list) should be a list of tuples (src_idx, vertno), and all locations
    should be present in the src.

    Parameters
    ----------
    location: np.array or callable
        The user input for the location.
    location_params: dict, optional
        Additional parameters to the location function, if needed.
    src: mne.SourceSpaces
        A source space that is used to validate the location function.

    Returns
    -------
    location: np.array or functools.partial
        Checked location list or function (partial object which does not
        require additional arguments anymore).
    n_vertices: the number of vertices
        The number of vertices that are created.

    Raises
    ------
    ValueError
        If any violation of the expected format is detected.
    Exception
        If any exception occurs during the function call.
    """

    vertices = location
    if callable(location):
        location = partial(location, **location_params)
        vertices = check_callable("location", location, src)

    check_vertices_list_of_tuples(vertices)
    check_vertices_in_src(vertices, src)

    return location, len(vertices)


def check_waveform(waveform, waveform_params, n_sources):
    """
    Check the user input for the waveforms of source activity.
    If waveform is a callable, it should not lead to an error.
    The result of the waveform function (if callable) or the waveform itself
    (if an array) should have the number of rows equal to the number of sources.

    Parameters
    ----------
    waveform: array or callable
        User input for the waveform.
    waveform_params: dict, optional
        Additional arguments for the waveform function, if needed.
    n_sources: int
        Number of expected sources. This number should be derived based on the
        provided input for location.

    Returns
    -------
    waveform: np.array or functools.partial
        Checked waveform array or function (partial object which does not
        require additional arguments anymore).

    Raises
    ------
    ValueError
        If any violation of the expected array shape is detected.
    Exception
        If any exception occurs during the function call.
    """

    data = waveform
    if callable(waveform):
        n_samples = 1000
        times = np.arange(n_samples) / n_samples
        waveform = partial(waveform, **waveform_params)
        data = check_callable("waveform", waveform, n_sources, times)

    if data.shape[0] != n_sources:
        raise ValueError(
            f"The number of sources in the provided array or in the result of"
            f"the provided function for source waveform does not match: "
            f"expected {n_sources}, got {data.shape[0]}"
        )

    if callable(waveform) and data.shape[1] != n_samples:
        raise ValueError(
            f"The number of samples in the result of the provided function"
            f"for source waveform does not match: expected {n_samples}, "
            f"got {data.shape[1]}"
        )

    return waveform


def check_names(names, n_sources, existing):
    """
    Check the user input for source names.
    The number of names should match the number of sources to be defined.
    In addition, all names should be unique and non-empty strings,
    which don't start with auto and are not already in the structure.

    Parameters
    ----------
    names: list
        The list of names to be added
    n_sources: int
        The number of sources to be added.
    existing: list
        The list of names which are already assigned to other sources.

    Raises
    ------
    ValueError
        If any of the aforementioned checks fail.
    """

    # Check the number of the provided names
    if len(names) != n_sources:
        raise ValueError(
            "The number of provided source names does not match "
            "the number of defined sources"
        )

    # All names should be non-empty strings
    for name in names:
        if not isinstance(name, str):
            actual_type = type(name).__name__
            raise ValueError(
                f"Expected all names to be strings, got {actual_type}: {name}"
            )

        if not name:
            raise ValueError("All names should not be empty")

        if name.startswith("auto"):
            raise ValueError(
                f"Name {name} should not start with auto, this prefix "
                f"is reserved for autogenerated names"
            )

        if name in existing:
            raise ValueError(f"Name {name} is already taken by another source")

    # Check that all names are unique
    if len(names) != len(set(names)):
        raise ValueError("All names should be unique")


def check_numeric_array(
    context, value, n_sources, bounds=(None, None), allow_none=False
):
    """
    Check the user input in case a list of numeric values for several sources
    is expected. If allowed, it can be None. Otherwise, a single float value
    that applies to all sources or an array of values with one for each source
    are expected.

    Parameters
    ----------
    context : str
        Context information for the error message.
    value : None, float, or array
        The provided value(s).
    n_sources : int
        The number of sources.
    bounds : tuple
        Bounds for the provided numeric values.
    allow_none : bool
        Whether None value is allowed.

    Raises
    ------
    ValueError
        If the provided value(s) do not follow the format described above.
    """

    if value is None and allow_none:
        return None

    values = np.ravel(np.array(value))
    if values.size != 1 and values.size != n_sources:
        raise ValueError(
            f"Expected either one {context} value that applies to all sources or "
            f"one {context} value for each of the {n_sources} sources, got {values.size}"
        )

    # Check that values are numeric and, if required, in bounds
    for value in values:
        check_numeric(context, value, bounds=bounds, allow_none=allow_none)

    # Broadcast to all sources if a single value was provided
    if values.size == 1:
        values = np.tile(values, (n_sources,))

    return values


def check_snr_params(snr_params, snr):
    """
    Check the user input for SNR parameters: if the SNR is adjusted (i.e., not None),
    then fmin and fmax should be present in the dictionary to define a frequency band.

    Parameters
    ----------
    snr_params: dict
        The provided dictionary with parameters of the SNR adjustment.
    snr: None, float, or array
        The provided value for SNR

    Raises
    ------
    ValueError
        If the provided snr_params dictionary does not have the necessary parameters.
    """
    if snr is None:
        return snr_params

    if "fmin" not in snr_params or "fmax" not in snr_params:
        raise ValueError(
            "Frequency band limits are required for the adjustment of SNR. "
            "Please add fmin and fmax to the snr_params dictionary."
        )

    if snr_params["fmin"] < 0 or snr_params["fmax"] < 0:
        raise ValueError("Frequency limits should be positive")

    return snr_params


def check_if_source_exists(name, existing, context):
    """
    Check if a source exists when trying to set the coupling.

    Parameters
    ----------
    name: str
        The name of the source to be checked.
    existing: list of str
        The name of all existing sources.
    context: str
        Context for an informative error message.

    Raises
    ------
    ValueError
        If the provided source name is not in the list of existing ones.
    """
    if name not in existing:
        raise ValueError(f"Source {name} {context}")


def check_coupling_params(method, coupling_params, coupling_edge):
    """
    Check whether all required coupling parameters were provided for the
    selected method.

    Parameters
    ----------
    method: str
        The name of the coupling method.
    coupling_params: dict
        The coupling parameters for the selected method.
    coupling_edge: tuple
        The coupling edge that the provided parameters apply to.
        It is only used to be more specific in the error message.

    Raises
    ------
    ValueError
        If the provided dictionary does not contain all required parameters.
    """

    # Test on a 10 second segment of white noise
    sfreq = 500
    rng = np.random.default_rng(seed=0)
    waveform = rng.random((10 * sfreq,))

    # Temporarily remove 'method' from coupling_params
    test_params = coupling_params.copy()
    test_params.pop("method")

    check_callable(
        f"coupling method, edge {coupling_edge}", method, waveform, sfreq, **test_params
    )


def check_coupling(coupling_edge, coupling_params, common_params, names, current_graph):
    """
    Check whether the provided coupling edge and parameters are valid.

    Parameters
    ----------
    coupling_edge: tuple
        The coupling edge (source, target) that the provided parameters apply to.
    coupling_params: dict
        The coupling parameters that were defined for this edge specifically.
    common_params: dict
        The coupling parameters that apply to all edges.
    names: list of str
        The names of sources that exist in the simulation.
    current_graph: nx.Graph
        The coupling graph that was already defined in the simulation

    Raises
    ------
    ValueError
        If source or target do not exist in the simulation.
        If the coupling edge was defined previously.
        If the coupling method or any of the required parameters for the method
        are not provided.
    """

    # Check that the coupling edge is defined as a tuple of two elements
    if not isinstance(coupling_edge, tuple):
        raise ValueError(f"Coupling edges {coupling_edge} should be defined as a tuple")
    if len(coupling_edge) != 2:
        raise ValueError(
            f"Coupling edges should contain two elements (names of "
            f"the source and the target), got {coupling_edge}"
        )

    # Check that both source names already exist
    source, target = coupling_edge
    check_if_source_exists(source, names, context="was not defined yet")
    check_if_source_exists(target, names, context="was not defined yet")

    # Check that the edge is not a self-loop
    if source == target:
        raise ValueError(
            f"The coupling edge {coupling_edge} is a self-loop, and "
            f"only connections between distinct sources are allowed."
        )

    # Check that this coupling edge has not been already added
    if current_graph.has_edge(*coupling_edge):
        raise ValueError(
            f"The coupling edge {coupling_edge} already exists in the "
            f"simulation, and multiple definitions are not allowed."
        )

    # Coupling parameters should be provided in a dictionary
    if not isinstance(coupling_params, dict):
        actual_type = type(coupling_params).__name__
        raise ValueError(
            f"Coupling parameters should be provided as a dictionary, "
            f"got {actual_type} for edge {coupling_edge}"
        )

    # Warn the user if some parameters were defined both for specific edges
    # and as common ones
    double_definition = set(common_params.keys()) & set(coupling_params.keys())
    if double_definition:
        double_defined = ", ".join(double_definition)
        warnings.warn(
            f"Parameters {double_defined} have double definition for edge "
            f"{coupling_edge}. Edge-specific values have higher priority."
        )

    # Overwrite the common coupling parameters with edge-specific ones
    params = common_params.copy()
    params.update(coupling_params)

    # Check that the coupling method was defined
    if "method" not in params:
        raise ValueError(
            f"Coupling method was not defined for the edge {coupling_edge}"
        )
    method = params["method"]

    # Check that the coupling method is a callable
    if not callable(method):
        raise ValueError(
            f"Expected coupling method to be a callable, "
            f"got {type(method).__name__} for edge {coupling_edge}"
        )

    # Check that all required coupling parameters were specified for the selected method
    check_coupling_params(method, params, coupling_edge)

    return params


def check_extents(extents, n_sources):
    # check if extents is a list, otherwise make it a list
    if not isinstance(extents, list):
        extents = [extents]
    # if extent is single number, propagate it to all patch sources
    if len(extents) == 1:
        extents = extents * n_sources

    for extent in extents:
        if extent is not None:
            # Check if each extent is a number
            if not isinstance(extent, (int, float, np.integer, np.floating)):
                raise ValueError(f"Extent {extent} must be a number.")

            # Check if each extent is positive
            if extent <= 0:
                raise ValueError(f"Extent {extent} must be a positive number.")

            # Issue a warning if any extent exceeds 1000 mm
            if extent > 1000:
                warnings.warn(
                    f"The extent {extent} (radius in mm) is more than 1000 mm. "
                    "Are you sure that the patch is supposed to be that big?",
                    UserWarning,
                )

    return extents


def check_stc_as_param(stc, src):
    for src_idx, s in enumerate(src):
        common = np.intersect1d(stc.vertices[src_idx], s["vertno"], assume_unique=True)

        # XXX: the code below overlaps with sources._BaseSource.check_compatibility
        missing_vertno = set(s["vertno"]) - set(common)
        if missing_vertno:
            report_missing = ", ".join([str(v) for v in missing_vertno])
            raise ValueError(
                f"The provided stc does not contain all vertices of the "
                f"source space that is used for simulations. The following vertices "
                f"from src[{src_idx}] are missing: {report_missing}"
            )


def check_colors(colors):
    """
    Check the dictionary with colors provided for the visualization of the
    source configuration.

    Parameters
    ----------
    color: dict
        The colors for different source types.

    Raises
    ------
    ValueError
        If the source type is not recognized or the color value is invalid.
    """
    if colors is None:
        return

    for key, color in colors.items():
        if key not in VIZ_SOURCE_TYPES:
            raise ValueError(f"Unexpected source type: {key}")

        # We expect a colormap for patches (easier for now but later we can
        # consider constructing a colormap from the provided color)
        if key == "patch" and color not in plt.colormaps():
            raise ValueError(
                "Expected a valid matplotlib colormap for the patch sources"
            )

        # For all other source types, one color is expected
        if key != "patch" and not is_color_like(color):
            raise ValueError(f"Expected a valid matplotlib color for the {key} sources")


def check_scale_factors(scale_factors):
    """
    Check the dictionary with scale factors provided for the visualization of the
    source configuration.

    Parameters
    ----------
    scale_factors: dict
        The scale factors for different source types.

    Raises
    ------
    ValueError
        If the source type is not recognized or the scale factor is negative or not
        numeric.
    """
    if scale_factors is None:
        return

    for key, scale_factor in scale_factors.items():
        if key not in VIZ_SOURCE_TYPES:
            raise ValueError(f"Unexpected source type: {key}")

        check_numeric(
            f"scale_factor of {key} sources", scale_factor, (0, None), allow_none=False
        )



import numpy as np
import mne

from meegsim._check import check_numeric, check_if_source_exists
from meegsim.sensor_noise import _adjust_sensor_noise, _prepare_sensor_noise
from meegsim.sources import _combine_sources_into_stc
from meegsim.viz import plot_source_configuration


class SourceConfiguration:
    """
    This class describes a simulated configuration of sources
    of brain activity and noise.

    Attributes
    ----------
    src : SourceSpaces
        Source spaces object that stores all candidate source locations.

    sfreq : float
        Sampling frequency of the simulated data, in Hz.

    duration : float
        Length of the simulated data, in seconds.

    random_state : int or None, optional
        Random state that was used to generate the configuration.
    """

    def __init__(self, src, sfreq, duration, random_state=None):
        self.src = src

        # Simulation parameters
        self.sfreq = sfreq
        self.duration = duration
        self.n_samples = self.sfreq * self.duration
        self.times = np.arange(self.n_samples) / self.sfreq
        self.tstep = self.times[1] - self.times[0]

        # Random state (for reproducibility & for sensor noise generation here)
        self.random_state = random_state

        # Keep track of all added sources, store 'signal' and 'noise' separately to ease the calculation of SNR
        self._sources = {}
        self._noise_sources = {}

    def __getitem__(self, name):
        """
        This function provides quick access to the simulated sources. The syntax
        is similar to list/dict access: ``sc[name]`` returns the corresponding source
        if it is present in the configuration.

        Parameters
        ----------
        name : str
            Name of the source.

        Returns
        -------
        source : PointSource or PatchSource
            The corresponding source.

        Raises
        ------
        ValueError
            In case there is no source with the provided name.
        """
        check_if_source_exists(
            name,
            list(self._sources.keys()),
            context="does not exist in the configuration",
        )
        return self._sources[name]

    def plot(
        self,
        subject,
        hemi="lh",
        colors=None,
        scale_factors=None,
        show_noise_sources=True,
        show_candidate_locations=False,
        **plot_kwargs,
    ):
        """
        Plot the source configuration. This function is built on top of
        :py:meth:`mne.SourceEstimate.plot`, and the meaning of all
        parameters is kept the same unless mentioned.

        Parameters
        ----------
        subject : str
            The name of the subject.
        hemi : str
            The hemisphere to plot. The values "both" and "split" are also
            supported.
        colors : dict or None, optional
            This dictionary can be used to override the default color palette,
            which is described in the Notes section below.
        scale_factors : dict or None, optional
            This dictionary can be used to override the default display size
            for sources of different types. See Notes for the default values.
        show_noise_sources : bool, optional
            Controls whether noise sources should be displayed (True by default).
        show_candidate_locations : bool, optional
            Controls whether candidate source locations (all vertices present in
            the source space) should be displayed (False by default).
        **kwargs: dict, optional
            Additional parameters that should be passed to
            :meth:`mne.SourceEstimate.plot`.

            .. warning::
                Plotting all candidate locations might require lots of resources
                depending on the size of the source space.

        Returns
        -------
        figure : mne.viz.Brain or matplotlib.figure.Figure
            The resulting figure.

        Notes
        -----
        The following parameters are used for plotting by default:

        .. code-block::

            DEFAULT_COLORS = dict(
                point="green",
                patch="Oranges",
                noise="black",
                candidate="yellow"
            )

            DEFAULT_SCALE_FACTORS = dict(
                point=0.75,
                noise=0.3,
                candidate=0.05
            )

            DEFAULT_PLOT_KWARGS = dict(
                background="w",
                cortex="low_contrast",
                colorbar=False,
                clim=dict(kind="value", lims=[0, 0.5, 1]),
                transparent=True,
            )
        """
        return plot_source_configuration(
            self,
            subject=subject,
            hemi=hemi,
            colors=colors,
            scale_factors=scale_factors,
            show_noise_sources=show_noise_sources,
            show_candidate_locations=show_candidate_locations,
            **plot_kwargs,
        )

    def to_stc(self):
        """
        Obtain an ``stc`` object that contains data from all sources
        in the configuration.

        Returns
        -------
        stc : SourceEstimate
            The resulting stc object that contains data from all sources.
        """
        sources = list(self._sources.values())
        noise_sources = list(self._noise_sources.values())
        all_sources = sources + noise_sources

        if not all_sources:
            raise ValueError("No sources were added to the configuration.")

        return _combine_sources_into_stc(all_sources, self.src, self.tstep)

    def to_raw(self, fwd, info, sensor_noise_level=None):
        """
        Project the activity of all simulated sources to sensor space.

        Parameters
        ----------
        fwd : Forward
            The forward model.
        info : Info
            The info structure that describes the channel layout.
        sensor_noise_level : float, optional
            The desired level of sensor-space noise between 0 and 1. For example,
            if 0.1 is specified, 10% of total sensor-space power will stem from
            white noise with an identity covariance matrix, while the remaining 90%
            of power will be explained by source activity projected to sensor space.
            By default, no sensor space noise is added. See Notes for more details.

        Returns
        -------
        raw : Raw
            The simulated sensor space data.

        Notes
        -----
        The adjustment of sensor space noise is performed as follows:

        1. The sensor space noise is scaled to equalize the mean sensor-space variance
        of broadband noise and brain activity.

        2. The brain activity and noise are mixed to achieve the desired level of
        sensor space noise (denoted by :math:`\\gamma` below):

        .. math::

            y = \\sqrt{1 - \\gamma} \\cdot y_{brain} + \\sqrt{\\gamma} \\cdot y_{noise}

        If the sensor noise is independent from projected brain activity, the following
        relationship will hold for the total sensor space power:

        .. math::

            P_{total} = (1 - \\gamma) \\cdot P_{brain} + \\gamma \\cdot P_{noise}
        """
        check_numeric("sensor_noise_level", sensor_noise_level, [0.0, 1.0])

        # Project source activity to sensor space
        stc_combined = self.to_stc()
        raw = mne.apply_forward_raw(fwd, stc_combined, info)

        # Add sensor space noise if needed
        if sensor_noise_level:
            noise = _prepare_sensor_noise(raw, self.times, self.random_state)
            raw = _adjust_sensor_noise(raw, noise, sensor_noise_level)

        return raw
